use std::ops::Deref;
use std::path::Path;

use async_std::sync::{Arc, Mutex};
use eyre::Result;
use tide::prelude::*;

mod state;
mod templates;
use state::State;

#[async_std::main]
async fn main() -> Result<()> {
    tide::log::start();

    let matches = clap::App::new("gityeet")
        .arg(clap::Arg::with_name("root")
            .help("sets the directory from which repositories will be found")
            .required(true))
        .arg(clap::Arg::with_name("port")
            .help("sets the port number")
            .required(true))
        .get_matches();

    let root = matches.value_of("root").expect("root is required");
    let port = matches.value_of("port").expect("port is required");
    let port: u16 = port.parse().expect("port was not valid");

    let state = State::discover(root).await?;
    // unfortunately, the State has to be both Send and Sync, but git2::Repository is not Sync
    // and even an Arc<RwLock<State>> isn't Sync if State isn't Sync, it has to be a Mutex to get Sync
    // TODO do something smarter for that like maybe just a vec of repo paths
    let mut app = tide::with_state(Arc::new(Mutex::new(state)));
    app.at("/").get(index);
    {
        let mut repo = app.at("/*repo/");
        repo.get(repo_index);
        repo.at("tree/*path").get(repo_tree_path);
    }
    app.listen(("127.0.0.1", port)).await?;
    Ok(())
}

type Request = tide::Request<Arc<Mutex<State>>>;

async fn index(req: Request) -> tide::Result {
    let state = req.state();
    let state = state.lock_arc().await;
    Ok(templates::Index { state: state.deref() }.into())
}

async fn repo_index(req: Request) -> tide::Result {
    let state = req.state();
    let repo_rel_path = req.param("repo").expect("no repo in repo-based URL?");

    view_tree_path(state, repo_rel_path, "").await
}

async fn repo_tree_path(req: Request) -> tide::Result {
    let state = req.state();
    let repo_rel_path = req.param("repo").expect("no repo in repo-based URL?");
    let tree_rel_path = req.param("path").expect("no path in path-based URL?");

    view_tree_path(state, repo_rel_path, tree_rel_path).await
}

async fn view_tree_path(state: &Arc<Mutex<State>>, repo_rel_path: &str, tree_rel_path: &str) -> tide::Result {
    let state = state.lock_arc().await;
    let repo = state.data.iter()
        .find(|repo| state.relative_path(repo.path()) == Path::new(repo_rel_path));
    let repo = match repo {
        Some(x) => x,
        None => return Ok("bruh that's not a real repo".into()),
    };

    let head = repo.head();
    let head = match head {
        Ok(x) => x,
        Err(err) => return Ok(format!("bruh that repo has no HEAD: {}", err).into()),
    };

    let head_tree = head.peel_to_tree();
    let head_tree = match head_tree {
        Ok(x) => x,
        Err(err) => return Ok(format!("bruh that repo's HEAD has no tree: {}", err).into()),
    };

    let selected_object = if tree_rel_path.is_empty() {
        head_tree.into_object()
    } else {
        let selected_tree_entry = head_tree.get_path(Path::new(&tree_rel_path));
        let selected_tree_entry = match selected_tree_entry {
            Ok(x) => x,
            Err(err) => return Ok(format!("bruh that path doesn't exist: {}", err).into()),
        };

        let selected_object = selected_tree_entry.to_object(repo);
        let selected_object = match selected_object {
            Ok(x) => x,
            Err(err) => return Ok(format!("bruh that path doesn't exist: {}", err).into()),
        };
        selected_object
    };

    let tree_rel_path = format!("/{}", tree_rel_path);

    if let Some(selected_tree) = selected_object.as_tree() {
        let template = templates::RepoFolder {
            repo,
            repo_path: repo_rel_path,
            title: repo_rel_path,
            rel_path: &tree_rel_path,
            tree: selected_tree.clone(),
        };
        Ok(template.into())
    } else if let Some(selected_blob) = selected_object.as_blob() {
        todo!()
    } else {
        panic!()
    }
}
